
__module_name__ = "_support_functions.py"
__doc__ = """Support module."""
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])
__version__ = "0.0.1"


# -- import packages: --------------------------------------------------------------------
from itertools import groupby
from typing import Union, Any
import numpy as np


# -- Supporting functions: ---------------------------------------------------------------
def as_list(input: Union[list, Any]):
    """Convert to list, if not already"""
    if isinstance(input, list):
        return input
    return [input]


def is_uniform(iterable):
    """Evaluate if all items in a list are uniform"""
    grouped = groupby(iterable)
    return next(grouped, True) and not next(grouped, False)


def power_space(start: int, stop: int, n: int, power: Union[int, float]):
    """
    Return integered-powered space.

    Parameters:
    -----------
    start
        first term of the power-space array.
        type: int

    stop
        final term of the power-space array.
        type: int

    n
        type: int
        length of array to be created

    power
        type: Union[int, float]
        power at which the space should decay / expand.

    Returns:
    --------
    power-spaced array
        type: np.ndarray
    """
    start_ = np.power(start, 1 / float(power))
    stop_  = np.power(stop, 1 / float(power))

    pspace = np.power(np.linspace(start_, stop_, num=n), power).astype(int)
    pspace[0], pspace[-1] = start, stop
    
    return pspace


def define_structure(in_features: int, out_features: int, hidden: Union[int, list]):
    
    """Build layered neural network structure"""
    hidden = as_list(hidden)
    n_hidden = len(hidden)

    layer_names = ["hidden_{}".format(i + 1) for i in range(n_hidden)] + ["output"]
    structure = [in_features] + hidden + [out_features]

    TorchNetDict = {}

    for n, (i, j) in enumerate(zip(structure[:-1], structure[1:])):
        TorchNetDict[layer_names[n]] = (i, j)

    return TorchNetDict




__module_name__ = "_layer.py"
__doc__ = """Layer module."""
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu"])
__version__ = "0.0.1"

# -- import packages: --------------------------------------------------------------------
from abc import ABC
from collections import OrderedDict
import torch


# -- Base Layer module: ------------------------------------------------------------------
class Layer(ABC):
    __name__ = ""

    def __init__(
        self,
        in_features: int,
        out_features: int,
        activation: torch.nn.modules.activation = None,
        bias: bool = True,
        dropout: float = 0,
        name: str = None,
    ):
        """
        
        Parameters:
        -----------
        in_features
            Size of layer input.
            type: int
        
        out_features
            Size of layer output.
            type: int
        
        activation
            If passed, defines appended activation function.
            type: 'torch.nn.modules.activation.<func>'
            default: None
        
        bias
            Indicate if the layer should not learn an additive bias.
            type: bool
            default: True
            
        dropout
            If > 0, append dropout layer with probablity p, where p = dropout.
            type: float
            default: 0
        
        name [Optional]
            Name of the layer. If defined, specified within torch.nn.Sequential.
            Layer name.
            type: str
                    
        Returns:
        --------
        None, instantiates Layer class.
        
        Notes:
        ------
        (1) General flow assumed is: Linear -> Dropout -> Activation
        """
        
        self.__parse_kwargs__(name, locals())

    # --- utilities: ---------------------------------------------------------------------
    def __configure_activation__(self, func):
        """Register the passed activation function."""
        if isinstance(func, str):
            return getattr(torch.nn, func)()
        if isinstance(func, torch.nn.Module):
            return func
        if isinstance(func(), torch.nn.Module):
            return func()
        
        msg = "Must pass torch.nn.<function> or a string that fetches a torch.nn.<function>"
        print(msg)

    def __parse_kwargs__(self, name, kwargs, ignore=["self", "name"]):
        """Register passed keyword arguments. Called on self.__init__()"""
        setattr(self, "__name__", name)

        self._kwargs = {}
        for k, v in kwargs.items():
            if k and (not k in ignore):
                self._kwargs[k] = v
                setattr(self, "_{}".format(k), v)

    # -- core properties: ----------------------------------------------------------------
    @property
    def linear(self):
        """torch.nn.Linear layer"""
        return torch.nn.Linear(
            in_features=self._in_features,
            out_features=self._out_features,
            bias=self._bias,
        )

    @property
    def dropout(self):
        """torch.nn.Dropout layer."""
        if self._dropout:
            return torch.nn.Dropout(self._dropout)

    @property
    def activation(self):
        """torch.nn.<activation> layer"""
        if self._activation:
            return self.__configure_activation__(self._activation)

    # -- called: -------------------------------------------------------------------------
    def __collect_attributes__(self):
        """Collect passed layer and optionally dropout, activation."""
        attributes = [i for i in self.__dir__() if not i.startswith("_")]
        for attr in attributes:
            if not getattr(self, attr) is None:
                if self.__name__:
                    attr_name = "_".join([self.__name__, attr])
                else:
                    attr_name = attr
                yield (attr_name, getattr(self, attr))

    def __call__(self)-> torch.nn.Sequential:
        """
        Generate layer from arguments passed to __init__() and processed with supporting
        functions.
        
        Returns:
        --------
        Layer
            Composed layer
            type: torch.nn.Sequential
        """
        return torch.nn.Sequential(OrderedDict(self.__collect_attributes__()))




__module_name__ = "_torch_net.py"
__doc__ = """Main user-facing API for torch.nn.Sequential."""
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu"])
__version__ = "0.0.1"


# -- import packages: --------------------------------------------------------------------
from collections import OrderedDict
from itertools import groupby
from typing import Union, Any
import torch


# -- import local dependencies: ----------------------------------------------------------
from .core._layer import Layer
from .core._support_functions import define_structure
from .core._base_torch_net import BaseTorchNet


# -- Main module class: ------------------------------------------------------------------
class TorchNet(BaseTorchNet):
    def __build__(self):

        TorchNetStructure = define_structure(
            self.in_features, self.out_features, self.hidden
        )

        TorchNetDict = OrderedDict()
        for n, (name, layer_dims) in enumerate(TorchNetStructure.items()):
            if name != "output":
                TorchNetDict[name] = Layer(
                    in_features=layer_dims[0],
                    out_features=layer_dims[1],
                    activation=self.activation[n],
                    bias=self.bias[n],
                    dropout=self.dropout[n],
                )()
            else:
                TorchNetDict[name] = Layer(
                    in_features=layer_dims[0],
                    out_features=layer_dims[1],
                    bias=self.output_bias,
                )()

        return TorchNetDict
    


# -- Main API-facing function: ----------------------------------------------
def _torch_net(
    in_features: int,
    out_features: int,
    hidden: Union[list, int] = [],
    activation="LeakyReLU",
    dropout: Union[float, list] = 0.2,
    bias: bool = True,
    output_bias: bool = True,
):
    """
    Parameters:
    -----------
    in_features
        Size of layer input.
        type: int

    out_features
        Size of layer output.
        type: int

    hidden
        list of hidden layer sizes
        type: Union[list, int]

    activation
        If passed, defines appended activation function.
        type: 'torch.nn.modules.activation.<func>'
        default: None

    dropout
        If > 0, append dropout layer with probablity p, where p = dropout.
        type: float
        default: 0

    bias
        Indicate if the layer should/should not learn an additive bias.
        type: bool
        default: True

    output_bias
        Indicate if the output layer should/should not learn an
        additive bias.
        type: bool
        defualt: True


    Returns:
    --------
    TorchNet
        Neural network block. To be accepted into a torch.nn.Module object.
        type: torch.nn.Sequential

    Notes:
    ------
    (1) For params: 'activation', 'bias', and 'dropout': if more
        params than necessary are passed, they will go unused.
    """

    return TorchNet(
        in_features=in_features,
        out_features=out_features,
        hidden=hidden,
        activation=activation,
        dropout=dropout,
        bias=bias,
        output_bias=output_bias,
    )()




__module_name__ = "_base_torch_net.py"
__doc__ = """Central module for torch-nets base class."""
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])
__version__ = "0.0.1"


# -- import packages: --------------------------------------------------------------------
from abc import ABC, abstractmethod
import torch


# -- import local dependencies: ----------------------------------------------------------
from ._support_functions import as_list, is_uniform


# -- supporting functions: ---------------------------------------------------------------
def format_layer_args(layer_arg, n_hidden):
    """format layer arguments w.r.t. the # of hidden layers"""
    layer_args = as_list(layer_arg)
        
    if len(layer_args) == 0:
        return layer_args

    if is_uniform(layer_args):
        return [layer_args[0]] * n_hidden

    if len(layer_args) != n_hidden:
        n_missing = n_hidden - len(layer_args)
        return layer_args + [layer_args[-1]] * n_missing

    else:
        return layer_args


# -- Base class: -------------------------------------------------------------------------
class BaseTorchNet(ABC):
    def __init__(
        self,
        in_features,
        out_features,
        hidden,
        activation="LeakyReLU",
        dropout=0.2,
        bias=True,
        output_bias=True,
    ):

        self.__parse__(locals())

    def __parse__(
        self,
        kwargs: dict,
        mk_list: list = ["hidden", "activation", "bias", "dropout"],
        ignore=["self"],
    ):

        ## should have one of these two scenarios
        if "hidden" in kwargs.keys():
            self.n_hidden = len(as_list(kwargs["hidden"]))
        else:
            self.n_hidden = kwargs["n_hidden"]
        ## --------------------------------------

        for key, val in kwargs.items():
            if not key in ignore:
                if key in mk_list:
                    val = format_layer_args(kwargs[key], n_hidden=self.n_hidden)
                setattr(self, key, val)

    @abstractmethod
    def __build__(self):
        """Must return OrderedDict"""

    def __call__(self):
        return torch.nn.Sequential(self.__build__())